#include <stdio.h>
#include <unistd.h>

#include "options.h"
#include "util.h"
#include "session.h"
#include "buffer.h"
#include "ssh.h"
#include "packet.h"
#include "auth.h"
#include "authpasswd.h"
#include "authpubkey.h"
#include "pwd.h"

static void authclear();
static int checkusername(unsigned char *username, unsigned int userlen);

void authinitialise() {

	/* TODO set tries left, etc */
	ses.authstate.failcount = 0;
	authclear();
	
}

static void authclear() {
	
	ses.authstate.authdone = 0;
	ses.authstate.pw = NULL;
	ses.authstate.authtypes = 0;
#ifdef MSSH_PUBKEY_AUTH
	ses.authstate.authtypes |= AUTH_TYPE_PUBKEY;
#endif
#ifdef MSSH_PASSWORD_AUTH
	ses.authstate.authtypes |= AUTH_TYPE_PASSWORD;
#endif

}

void recv_msg_userauth_request() {

	unsigned char *username, *servicename, *methodname;
	unsigned int userlen, servicelen, methodlen;

	TRACE("enter recv_msg_userauth_request");

	/* ignore packets if auth is already done */
	if (ses.authstate.authdone == 1) {
		return;
	}
	
	username = buf_getstring(ses.payload, &userlen);
	servicename = buf_getstring(ses.payload, &servicelen);
	methodname = buf_getstring(ses.payload, &methodlen);

	/* only handle 'ssh-connection' currently */
	if (servicelen != SSH_SERVICE_CONNECTION_LEN
			&& (strncmp(servicename, SSH_SERVICE_CONNECTION,
					SSH_SERVICE_CONNECTION_LEN) != 0)) {
		
		/* TODO - disconnect here */
		m_free(username);
		m_free(servicename);
		m_free(methodname);
		mssh_error("unknown service in auth");
	}

	printf("user = '%s' service = '%s' method = '%s'\n",
			username, servicename, methodname);

	if (methodlen == AUTH_METHOD_NONE_LEN &&
			strncmp(methodname, AUTH_METHOD_NONE,
				AUTH_METHOD_NONE_LEN) == 0) {
		send_msg_userauth_failure(0, 0);
	} else if (checkusername(username, userlen) == CHECK_USER_RETURN) {
		/* username is invalid/no shell/etc - send failure */
		TRACE("sending checkusername failure");
		send_msg_userauth_failure(0, 1);
#ifdef MSSH_PASSWORD_AUTH
	} else if (methodlen == AUTH_METHOD_PASSWORD_LEN &&
			strncmp(methodname, AUTH_METHOD_PASSWORD,
				AUTH_METHOD_PASSWORD_LEN) == 0) {
		passwordauth(username, userlen);
#endif /* MSSH_PASSWORD_AUTH */
#ifdef MSSH_PUBKEY_AUTH
	} else if (methodlen == AUTH_METHOD_PUBKEY_LEN &&
			strncmp(methodname, AUTH_METHOD_PUBKEY,
				AUTH_METHOD_PUBKEY_LEN) == 0) {
		pubkeyauth(username, userlen);
#endif /* MSSH_PUBKEY_AUTH */
	} else {
		send_msg_userauth_failure(0, 1);
	}

	m_free(username);
	m_free(servicename);
	m_free(methodname);
}

static int checkusername(unsigned char *username, unsigned int userlen) {

	char* shell;
	
	TRACE("enter checkusername");
	if (userlen > MAX_USERNAME_LEN) {
		mssh_error("username too long");
	}

	/* new user or username has changed */
	if (ses.authstate.pw == NULL ||
		strcmp(username, ses.authstate.pw->pw_name) != 0) {
			/* the username needs resetting */
			authclear();
			ses.authstate.pw = getpwnam((char*)username);
	}

	/* check that user exists */
	if (ses.authstate.pw == NULL) {
		TRACE("leave checkusername: user doesn't exist");
		send_msg_userauth_failure(0, 1);
		return CHECK_USER_RETURN;
	}

	/* check for an empty password */
	if (ses.authstate.pw->pw_passwd[0] == '\0') {
		TRACE("leave checkusername: empty pword");
		send_msg_userauth_failure(0, 1);
		return CHECK_USER_RETURN;
	}

#ifdef DEBUG_TRACE
	printf("shell is %s\n", ses.authstate.pw->pw_shell);
#endif
	/* check that the shell is valid */
	/* XXX - todo check this is correct: empty shell is ok */
	if (ses.authstate.pw->pw_shell[0] == '\0') {
		goto goodshell;
	}
	setusershell();
	while ((shell = getusershell()) != NULL) {
#ifdef DEBUG_TRACE
		printf("test shell is '%s'\n", shell);
#endif
		if (strcmp(shell, ses.authstate.pw->pw_shell) == 0) {
			/* have a match */
			goto goodshell;
		}
	}
	/* no matching shell */
	endusershell();
	TRACE("no matching shell");
	send_msg_userauth_failure(0, 1);
	return CHECK_USER_RETURN;
	
goodshell:
	endusershell();
	TRACE("matching shell");

	printf("uid = %d\n", ses.authstate.pw->pw_uid);
	TRACE("leave checkusername");
	return CHECK_USER_CONTINUE;

}

void send_msg_userauth_failure(int partial, int incrfail) {

	buffer *typebuf;

	TRACE("enter send_msg_userauth_failure");

	assert(ses.writepayload->len == 0 && ses.writepayload->pos == 0);
	
	buf_putbyte(ses.writepayload, SSH_MSG_USERAUTH_FAILURE);

	/* put a list of allowed types */
	typebuf = buf_new(30); /* long enough for PUBKEY and PASSWORD */

	if (ses.authstate.authtypes & AUTH_TYPE_PUBKEY) {
		buf_putbytes(typebuf, AUTH_METHOD_PUBKEY, AUTH_METHOD_PUBKEY_LEN);
		if (ses.authstate.authtypes & AUTH_TYPE_PASSWORD) {
			buf_putbyte(typebuf, ',');
		}
	}
	
	if (ses.authstate.authtypes & AUTH_TYPE_PASSWORD) {
		buf_putbytes(typebuf, AUTH_METHOD_PASSWORD, AUTH_METHOD_PASSWORD_LEN);
	}

	buf_setpos(typebuf, 0);
	buf_putstring(ses.writepayload, buf_getptr(typebuf, typebuf->len),
			typebuf->len);
	buf_free(typebuf);

	buf_putbyte(ses.writepayload, partial ? 1 : 0);
	encrypt_packet();

	if (incrfail) {
		ses.authstate.failcount++;
	}

	if (ses.authstate.failcount >= MAX_AUTH_TRIES) {
		/* XXX - send disconnect ? */
		fprintf(stderr, "Max auth tries reached, exiting");
		exit(0);
	}
	
	TRACE("leave send_msg_userauth_failure");
}

/* only to be called when writebuf is empty */
void send_msg_userauth_success() {

	TRACE("enter send_msg_userauth_success");

	assert(ses.writepayload->len == 0 && ses.writepayload->pos == 0);

	buf_putbyte(ses.writepayload, SSH_MSG_USERAUTH_SUCCESS);
	encrypt_packet();

	ses.authstate.authdone = 1;
	close(ses.childpipe);
	TRACE("leave send_msg_userauth_success");

}
