#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>

#include "options.h"
#include "session.h"
#include "packet.h"
#include "ssh.h"
#include "buffer.h"
#include "util.h"
#include "channel.h"
#include "chansession.h"
#include "ssh.h"

static struct Channel* newchannel(unsigned int remotechan, char type, 
		unsigned int transwindow, unsigned int transmaxpacket);
static void send_msg_channel_open_failure(unsigned int remotechan, int reason,
		const unsigned char *text, const unsigned char *lang);
static void send_msg_channel_open_success(struct Channel* channel,
		unsigned int recvwindow, 
		unsigned int recvmaxpacket);
static void writechannel(struct Channel *channel);
static void send_msg_channel_window_adjust(struct Channel *channel, 
		unsigned int incr);
static void send_msg_channel_data(struct Channel *channel, int isextended,
		unsigned int exttype);
static void send_msg_channel_eof(struct Channel *channel);
static void send_msg_channel_close(struct Channel *channel);
static void closechannel(struct Channel *channel);

/* initialise channels */
void chaninitialise() {

	/* may as well create space for a single channel */
	ses.channels = (struct Channel**)m_malloc(sizeof(struct Channel*));
	ses.chansize = 1;
	ses.channels[0] = NULL;

	chansessinitialise();
}

void chancleanup() {

	int i;

	TRACE("enter chancleanup");
	for (i = 0; i < ses.chansize; i++) {
		if (ses.channels[i] != NULL) {
			fprintf(stderr, "channel %d closing\n", i);
			closechannel(ses.channels[i]);
		}
	}
	m_free(ses.channels);
	TRACE("leave chancleanup");
}
			

/* create a new channel entry, send a reply confirm or failure */
static struct Channel* newchannel(unsigned int remotechan, char type, 
		unsigned int transwindow, unsigned int transmaxpacket) {

	struct Channel ** chanlist;
	struct Channel * newchan;
	int i;

	TRACE("enter newchannel");
	
	/* first see if we can use existing channels */
	for (i = 0; i < ses.chansize; i++) {
		if (ses.channels[i] == NULL) {
			break;
		}
	}

	/* otherwise extend the list */
	if (i == ses.chansize) {
		if (ses.chansize > MAX_CHANNELS) {
			send_msg_channel_open_failure(remotechan,
					SSH_OPEN_RESOURCE_SHORTAGE, "", "");
			TRACE("enter newchannel: max chans reached");
			return NULL;
		}

		/* extend the channels */
		chanlist = (struct Channel**)m_realloc(ses.channels,
				(ses.chansize+CHAN_EXTEND_SIZE)*sizeof(struct Channel*));
		/*if (chanlist != NULL) {*/

		ses.channels = chanlist;
		ses.chansize += CHAN_EXTEND_SIZE;

	/*	} else {
			send_msg_channel_open_failure(remotechan,
					SSH_OPEN_RESOURCE_SHORTAGE, "", "");
			TRACE("enter newchannel: couldn't allocate more slots");
			return null;
		}*/
	}
	
	newchan = (struct Channel*)m_malloc(sizeof(struct Channel));
	newchan->type = type;
	newchan->index = i;
	newchan->sentclosed = 0;
	newchan->remotechan = remotechan;
	newchan->transwindow = transwindow;
	newchan->transmaxpacket = transmaxpacket;
	newchan->typedata = NULL;
	newchan->infd = -1;
	newchan->outfd = -1;

	newchan->writebuf = buf_new(RECV_MAXWINDOW);
	newchan->recvwindow = RECV_MAXWINDOW;
	newchan->recvmaxpacket = RECV_MAXPACKET;

	ses.channels[i] = newchan;

	send_msg_channel_open_success(newchan, RECV_MAXWINDOW,
			RECV_MAXPACKET);
	TRACE("enter newchannel");

	return newchan;
}

static struct Channel* getchannel(unsigned int chan) {
	if (chan >= ses.chansize || ses.channels[chan] == NULL) {
		return NULL;
	}
	return ses.channels[chan];
}

void channelio(fd_set *readfd, fd_set *writefd) {

	struct Channel *channel;
	int i;

	for (i = 0; i < ses.chansize; i++) {
		channel = ses.channels[i];
		
		if (channel != NULL) {
			/* read from program */
			if (channel->outfd != -1 &&
					FD_ISSET(channel->outfd, readfd)) {
				send_msg_channel_data(channel, 0, 0);
			}
			/* write to program */
			if (channel->infd != -1 &&
					FD_ISSET(channel->infd, writefd)) {
				writechannel(channel);
			}
			/* read from program stderr for interactive sessions */
			if (channel->type == CHANNEL_ID_SESSION) {
				int errfd = 
					((struct ChanSess *)channel->typedata)->errfd;
				if (errfd != -1 && FD_ISSET(errfd, readfd)) {
					send_msg_channel_data(channel, 1, SSH_EXTENDED_DATA_STDERR);
			
				}
			}
		}
	}
}

static void send_msg_channel_close(struct Channel *channel) {

	assert(ses.writepayload->len == 0 && ses.writepayload->pos == 0);

	buf_putbyte(ses.writepayload, SSH_MSG_CHANNEL_CLOSE);
	buf_putint(ses.writepayload, channel->remotechan);

	encrypt_packet();

	channel->sentclosed = 1;
}

static void send_msg_channel_eof(struct Channel *channel) {

	assert(ses.writepayload->len == 0 && ses.writepayload->pos == 0);

	buf_putbyte(ses.writepayload, SSH_MSG_CHANNEL_EOF);
	buf_putint(ses.writepayload, channel->remotechan);

	encrypt_packet();

}
/* only called when we know we can write to a channel, writes as much as
 * possible */
static void writechannel(struct Channel* channel) {

	int len, maxlen;
	buffer *buf;

	TRACE("enter writechannel");
	buf = channel->writebuf;
	maxlen = buf->len - buf->pos;

	len = write(channel->infd, buf_getptr(buf, maxlen), maxlen);
	if (len <= 0) {
		/* close the channel, we'll cleanup when they reply */
		if (errno != EINTR && !channel->sentclosed) {
			if (channel->type == CHANNEL_ID_SESSION) {
				struct ChanSess * chansess;
				chansess = (struct ChanSess*)channel->typedata;
				if (chansess->exited) {
					if (chansess->exitsignal > 0) {
						send_msg_chansess_exitsignal(channel, chansess);
					} else {
						send_msg_chansess_exitstatus(channel, chansess);
					}
				}
			}
		   send_msg_channel_close(channel);
		}
		return;
	}
	
	if (len == maxlen) {
		buf_setpos(buf, 0);
		buf_setlen(buf, 0);
		send_msg_channel_window_adjust(channel, buf->size
				- channel->recvwindow);
		channel->recvwindow = buf->size;
	} else {
		buf_incrpos(buf, len);
	}
	/* TODO - this is inefficient */
	TRACE("leave writechannel");
}

void setchannelfds(fd_set *readfd, fd_set *writefd) {
	
	int i;
	struct Channel * channel;
	
	for (i = 0; i < ses.chansize; i++) {
		channel = ses.channels[i];
		if (channel != NULL && channel->sentclosed != 1) {

			/* stdout and stderr */
			if (channel->transwindow > 0) {

				/* stdout */
				if (channel->outfd != -1) {
					/* there's space to read more from the program */
					FD_SET(channel->outfd, readfd);
				}
				/* stderr for interactive sessions */
				if (channel->type == CHANNEL_ID_SESSION) {
					int errfd = 
						((struct ChanSess *)channel->typedata)->errfd;
					if (errfd != -1) {
						FD_SET(errfd, readfd);
					}
				}
			}
			/* stdin */
			if (channel->infd != -1 &&
					channel->writebuf->pos 
						< channel->writebuf->len) {
				/* there's space to write more to the program */
				FD_SET(channel->infd, writefd);
			}
		}
	}
}

void recv_msg_channel_close() {

	unsigned int chan;
	struct Channel * channel;

	TRACE("enter recv_msg_channel_close");

	chan = buf_getint(ses.payload);
	channel = getchannel(chan);

	if (channel == NULL) {
		/* disconnect ? */
		mssh_error("Unknown channel");
	}
	assert(ses.authstate.authdone);

	if (!channel->sentclosed) {
		send_msg_channel_close(channel);
	}

	closechannel(channel);
	
	TRACE("leave recv_msg_channel_close");
}

static void closechannel(struct Channel * channel) {

	unsigned int index;

	TRACE("enter closechannel");
#ifdef DEBUG_TRACE
	fprintf(stderr, "channel index is %d\n", channel->index);
#endif
	
	buf_free(channel->writebuf);
	TRACE("frees done ");

	if (channel->type == CHANNEL_ID_SESSION) {
		closechansess(channel);
	}

	index = channel->index;
	m_free(channel);
	ses.channels[index] = NULL;
	TRACE("leave closechannel");
}

void recv_msg_channel_request() {

	unsigned int chan;
	struct Channel *channel;

	TRACE("enter recv_msg_channel_request");
	
	chan = buf_getint(ses.payload);
	channel = getchannel(chan);

	if (channel == NULL) {
		/* disconnect ? */
		mssh_error("Unknown channel");
	}

	/* if the channel exists, then we've checked for auth when creating */
	assert(ses.authstate.authdone);

#ifdef DEBUG_TRACE
	fprintf(stderr, "chan type is %d\n", channel->type);
#endif

	/* handle according to channel type */
	switch (channel->type) {

		case CHANNEL_ID_SESSION:
			TRACE("continue recv_msg_channel_request: session request");
			chansessionrequest(channel);
			break;

		case CHANNEL_ID_X11:

		default:
			send_msg_channel_failure(channel);
	}

	TRACE("leave recv_msg_channel_request");

}

/* chan is the remote channel, isextended is 0 if it is normal data, 1
 * if it is extended data. if it is extended, then the type is in
 * exttype */
static void send_msg_channel_data(struct Channel *channel, int isextended,
		unsigned int exttype) {

	buffer *buf;
	int len;
	unsigned int maxlen;
	int fd;

	TRACE("enter send_msg_channel_data");
#ifdef DEBUG_TRACE
	fprintf(stderr, "extended = %d type = %d\n", isextended, exttype );
#endif

	assert(ses.writepayload->len == 0 && ses.writepayload->pos == 0);
	
	maxlen = MIN(channel->transwindow, channel->transmaxpacket);
	/* -(1+4+4) is SSH_MSG_CHANNEL_DATA, channel number, string length, and 
	 * exttype if is extended */
	maxlen = MIN(maxlen, ses.writepayload->size 
			- 1 - 4 - 4 - (isextended ? 4 : 0));
	if (maxlen == 0) {
		TRACE("leave send_msg_channel_data: no window");
		return; /* the data will get written later */
	}

	if (isextended) {
		TRACE("isextended is true");
		fd = ((struct ChanSess*)channel->typedata)->errfd;
	} else {
		fd = channel->outfd;
	}

	assert(fd >= 0);

	/* read the data */
	buf = buf_new(maxlen);
	len = read(fd, buf_getwriteptr(buf, maxlen), maxlen);
	if (len <= 0) {
		/* close the channel, we'll cleanup when they reply */
		if (errno != EINTR && !channel->sentclosed) {
			if (channel->type == CHANNEL_ID_SESSION) {
				struct ChanSess * chansess;
				chansess = (struct ChanSess*)channel->typedata;
				if (chansess->exited) {
					if (chansess->exitsignal > 0) {
						send_msg_chansess_exitsignal(channel, chansess);
					} else {
						send_msg_chansess_exitstatus(channel, chansess);
					}
				}
			}
			
			send_msg_channel_eof(channel);
			send_msg_channel_close(channel);
		}
		buf_free(buf);
		TRACE("leave send_msg_channel_data: len <= 0");
		return;
	}
	buf_incrlen(buf, len);

	buf_putbyte(ses.writepayload, 
			isextended ? SSH_MSG_CHANNEL_EXTENDED_DATA : SSH_MSG_CHANNEL_DATA);
	buf_putint(ses.writepayload, channel->remotechan);

	if (isextended) {
		buf_putint(ses.writepayload, exttype);
	}

	buf_putstring(ses.writepayload, buf_getptr(buf, len), len);
	buf_free(buf);

	channel->transwindow -= len;

	encrypt_packet();
	TRACE("leave send_msg_channel_data");
}


/* when we receive channel data */
void recv_msg_channel_data() {

	unsigned int chan;
	struct Channel * channel;
	unsigned int datalen;
	unsigned int pos;

	TRACE("enter recv_msg_channel_data");
	
	chan = buf_getint(ses.payload);
	channel = getchannel(chan);
	if (channel == NULL) {
		/* disconnect ? */
		mssh_error("Unknown channel");
	}

	assert(ses.authstate.authdone);
	assert(channel->infd != -1);
	assert(channel->sentclosed == 0);

	datalen = buf_getint(ses.payload);

	/* if the client is going to send us more data than we've allocated, then 
	 * it has ignored the windowsize, so we "MAY ignore all extra data" */
	if (datalen > channel->writebuf->size - channel->writebuf->pos) {
		TRACE("Warning: recv_msg_channel_data: extra data past window");
		datalen = channel->writebuf->size - channel->writebuf->pos;
	}

	/* write to the buffer - we always append to the end of the buffer */
	pos = channel->writebuf->pos;
	buf_setpos(channel->writebuf, channel->writebuf->len);
	memcpy(buf_getwriteptr(channel->writebuf, datalen), 
			buf_getptr(ses.payload, datalen), datalen);
	buf_incrwritepos(channel->writebuf, datalen);
	buf_setpos(channel->writebuf, pos); /* revert pos */

	channel->recvwindow -= datalen;
/*	if (channel->recvwindow < RECV_MINWINDOW) {
		send_msg_channel_window_adjust(channel, 
				RECV_MAXWINDOW - channel->recvwindow);
		channel->recvwindow = RECV_MAXWINDOW;
	}*/

	TRACE("leave recv_msg_channel_data");
}

void recv_msg_channel_window_adjust() {

	unsigned int chan;
	struct Channel * channel;
	unsigned int incr;
	
	chan = buf_getint(ses.payload);
	channel = getchannel(chan);

	if (channel == NULL) {
		mssh_error("Invalid channel"); /* TODO - disconnect */
	}
	
	incr = buf_getint(ses.payload);
#ifdef DEBUG_TRACE
	fprintf(stderr, "window increment %d\n", incr);
#endif
#if 0
	if (incr > ses.channels[chan]->transmaxpacket) {
		mssh_error("Invalid increment"); /* TODO - disconnect */
	}
#endif
	channel->transwindow += incr;

}

static void send_msg_channel_window_adjust(struct Channel* channel, 
		unsigned int incr) {

#ifdef DEBUG_TRACE
	fprintf(stderr, "sending window adjust %d\n", incr);
#endif
	assert(ses.writepayload->len == 0 && ses.writepayload->pos == 0);

	buf_putbyte(ses.writepayload, SSH_MSG_CHANNEL_WINDOW_ADJUST);
	buf_putint(ses.writepayload, channel->remotechan);
	buf_putint(ses.writepayload, incr);

	encrypt_packet();
}
	
void recv_msg_channel_open() {

	unsigned char* type;
	unsigned int typelen;
	unsigned int remotechan, transwindow, transmaxpacket;
	struct Channel* channel;


	TRACE("enter recv_msg_channel_open");

	/* check we have auth */
	if (ses.authstate.authdone != 1) {
		mssh_error("Attempt to open channel before userauth");
	}
	
	type = buf_getstring(ses.payload, &typelen);

	remotechan = buf_getint(ses.payload);
	transwindow = buf_getint(ses.payload);
	transmaxpacket = buf_getint(ses.payload);
	transmaxpacket = MIN(transmaxpacket, MAX_TRANS_PAYLOAD_LEN);

	if (typelen > MAX_NAME_LEN) {
		/* send channel_open_failure below */
	} else if (strcmp(type, "session") == 0) {
			channel = newchannel(remotechan, CHANNEL_ID_SESSION, 
					transwindow, transmaxpacket);
			if (channel != NULL) {
				newchansess(channel);
			}
			goto out;
	} else if (strcmp(type, "x11") == 0) {
			newchannel(remotechan, CHANNEL_ID_X11,
					transwindow, transmaxpacket);
			goto out;
	}

	send_msg_channel_open_failure(remotechan, 
			SSH_OPEN_UNKNOWN_CHANNEL_TYPE,
			"Unknown channel", "en");
out:
	m_free(type);

	TRACE("leave recv_msg_channel_open");
}

void send_msg_channel_failure(struct Channel *channel) {

	TRACE("enter send_msg_channel_failure");
	assert(ses.writepayload->len == 0 && ses.writepayload->pos == 0);

	buf_putbyte(ses.writepayload, SSH_MSG_CHANNEL_FAILURE);
	buf_putint(ses.writepayload, channel->remotechan);

	encrypt_packet();
	TRACE("leave send_msg_channel_failure");
}

void send_msg_channel_success(struct Channel *channel) {

	TRACE("enter send_msg_channel_success");
	assert(ses.writepayload->len == 0 && ses.writepayload->pos == 0);

	buf_putbyte(ses.writepayload, SSH_MSG_CHANNEL_SUCCESS);
	buf_putint(ses.writepayload, channel->remotechan);

	encrypt_packet();
	TRACE("leave send_msg_channel_success");
}

static void send_msg_channel_open_failure(unsigned int remotechan, 
		int reason, const unsigned char *text, const unsigned char *lang) {

	TRACE("enter send_msg_channel_open_failure");
	assert(ses.writepayload->len == 0 && ses.writepayload->pos == 0);
	
	buf_putbyte(ses.writepayload, SSH_MSG_CHANNEL_OPEN_FAILURE);
	buf_putint(ses.writepayload, remotechan);
	buf_putint(ses.writepayload, reason);
	buf_putstring(ses.writepayload, text, strlen((char*)text));
	buf_putstring(ses.writepayload, lang, strlen((char*)lang));

	encrypt_packet();
	TRACE("leave send_msg_channel_open_failure");
}

static void send_msg_channel_open_success(struct Channel* channel,
		unsigned int recvwindow, 
		unsigned int recvmaxpacket) {

	TRACE("enter send_msg_channel_open_success");
	assert(ses.writepayload->len == 0 && ses.writepayload->pos == 0);

	buf_putbyte(ses.writepayload, SSH_MSG_CHANNEL_OPEN_CONFIRMATION);
	buf_putint(ses.writepayload, channel->remotechan);
	buf_putint(ses.writepayload, channel->index);
	buf_putint(ses.writepayload, recvwindow);
	buf_putint(ses.writepayload, recvmaxpacket);

	encrypt_packet();
	TRACE("leave send_msg_channel_open_success");
}
