#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "options.h"
#include "util.h"
#include "session.h"
#include "buffer.h"
#include "signkey.h"

static sign_key * loadhostkeys();
static void listensocket(int *sock);
static void sigchld_handler(int dummy);
static void sigsegv_handler(int);

static int childpipes[MAX_UNAUTH_CLIENTS];

int main(int argc, char ** argv) {
	
	fd_set fds;
	struct timeval seltimeout;
	int i, j, val;
	int maxsock;
	struct sockaddr_in remote_addr;
	int remote_addrlen;
	sign_key *hostkey;
	int listensocks[MAX_LISTEN_ADDR];
	unsigned int listensockcount = 0;

	int childsock;
	pid_t childpid;
	int childpipe[2];

	struct sigaction sa_chld;

#if 0 /* leave off for debugging */
	i = fork();
	if (i < 0) {
		mssh_error("Failed to create background process");
	}
	if (i > 0) {
		exit(0);
	}

	fprintf(stderr, "Running in background\n");
#endif

	
	/* load hostkey(s) */
	hostkey = loadhostkeys();
	
	/* setup the sockets */
	/* TODO make this configurable */
	listensocket(&listensocks[0]);
	listensockcount = 1;
	maxsock = listensocks[listensockcount-1];

	/* catch and reap zombie children */
	
	sa_chld.sa_handler = sigchld_handler;
	sa_chld.sa_flags = SA_NOCLDSTOP;
	if (sigaction(SIGCHLD, &sa_chld, NULL) < 0) {
		mssh_error("signal() error");
	}
	if (signal(SIGSEGV, sigsegv_handler) == SIG_ERR) {
		mssh_error("signal() error");
	}
		
	/* sockets to identify pre-authenticated clients */
	for (i = 0; i < MAX_UNAUTH_CLIENTS; i++) {
		childpipes[i] = -1;
	}

	/* incoming connection select loop */
	for(;;) {

		FD_ZERO(&fds);
		
		seltimeout.tv_sec = 60; /* TODO find good value */
		seltimeout.tv_usec = 0;
		
		/* listening sockets */
		for (i = 0; i < listensockcount; i++) {
			FD_SET(listensocks[i], &fds);
		}

		/* pre-authentication clients */
		for (i = 0; i < MAX_UNAUTH_CLIENTS; i++) {
			if (childpipes[i] >= 0) 
			FD_SET(childpipes[i], &fds);
			maxsock = MAX(maxsock, childpipes[i]);
		}

		val = select(maxsock+1, &fds, NULL, NULL, &seltimeout);
		printf("main select here <<<<<<==========\n");
		if (val == 0) {
			/* timeout reached */
			continue;
		}
		if (val < 0) {
			if (errno == EINTR) {
				TRACE("EINTR");
				continue;
			}
			mssh_error("listening socket error");
		}

		/* close fds which have been authed or closed */
		for (i = 0; i < MAX_UNAUTH_CLIENTS; i++) {
			if (childpipes[i] >= 0 && FD_ISSET(childpipes[i], &fds)) {
				fprintf(stderr, "closing childpipe %d\n", i);
				close(childpipes[i]);
				childpipes[i] = -1;
			}
		}

		/* handle each socket which has something to say */
		for (i = 0; i < listensockcount; i++) {
			if (!FD_ISSET(listensocks[i], &fds)) 
				continue;

			/* child connection */
			remote_addrlen = sizeof(struct sockaddr_in);
			childsock = accept(listensocks[i], 
					(struct sockaddr*)&remote_addr, &remote_addrlen);

			if (childsock < 0) {
				/* accept failed */
				continue;
			}

			/* check for max number of connections not authorised */
			for (j = 0; j < MAX_UNAUTH_CLIENTS; j++) {
				if (childpipes[j] < 0) {
					break;
				}
			}
			if (j == MAX_UNAUTH_CLIENTS) {
				/* no free connections */
				printf("no free connections\n");
				close(childsock);
				continue;
			}

			if (pipe(childpipe) < 0) {
				TRACE("error creating child pipe");
				close(childsock);
				continue;
			}

			if ((childpid = fork()) == 0) {

				/* child */
				if (setpgid(0,0) < 0) {
					mssh_error("child error");
				}

				for (i = 0; i < listensockcount; i++) {
					close(listensocks[i]);
				}
				close(childpipe[0]);
				/* start the session */
				child_session(childsock, hostkey, childpipe[1]);
				/* don't return */
				assert(0);
			}
			
			/* parent */
			childpipes[j] = childpipe[0];
			fprintf(stderr, "allocating child pipe %d\n", j);
			close(childpipe[1]);
			close(childsock);
		}
	} /* for(;;) loop */

	/* don't reach here */
	return -1;
}

/* catch + reap zombie children */
static void sigchld_handler(int fish) {
	struct sigaction sa_chld;

	while(waitpid(-1, NULL, WNOHANG) > 0); 

	sa_chld.sa_handler = sigchld_handler;
	sa_chld.sa_flags = SA_NOCLDSTOP;
	if (sigaction(SIGCHLD, &sa_chld, NULL) < 0) {
		mssh_error("signal() error");
	}
}

/* catch any segvs */
static void sigsegv_handler(int fish) {
	fprintf(stderr, "Aiee, segfault! You should probably report "
			"this as a bug\n");
	exit(1);
}
	
static void listensocket(int *sock) {
	
	int listensock; /* listening fd */
	struct sockaddr_in listen_addr;
	struct linger linger;
	int val;

	/* open a socket */
	listensock = socket(PF_INET, SOCK_STREAM, 0);
	if (listensock < 0) {
		mssh_error("socket() failed");
	}

	/* set to reuse, quick timeout */
	setsockopt(listensock, SOL_SOCKET, SO_REUSEADDR,
			(void*) &val, sizeof(val));
	linger.l_onoff = 1;
	linger.l_linger = 5;
	setsockopt(listensock, SOL_SOCKET, SO_LINGER,
			(void*)&linger, sizeof(linger));

	/* bind to any currently TODO config*/
	listen_addr.sin_family = AF_INET;
	listen_addr.sin_port = htons(MSSH_PORT);
	listen_addr.sin_addr.s_addr = INADDR_ANY;
	memset(&(listen_addr.sin_zero), '\0', 8); /* XXX neccesary? */

	if (bind(listensock, (struct sockaddr *)&listen_addr, 
				sizeof(struct sockaddr)) < 0) {
		mssh_error("bind() failed");
	}

	/* listen */
	if (listen(listensock, 20) < 0) { /* TODO set listen count */
		mssh_error("listen() failed");
	}

	/* nonblock */
	if (fcntl(listensock, F_SETFL, O_NONBLOCK) < 0) {
		mssh_error("setting O_NONBLOCK failed");
	}

	*sock = listensock;
}

static sign_key * loadhostkeys() {

	sign_key * hostkey;
	buffer *buf;

	TRACE("enter loadhostkeys");
	hostkey = new_sign_key();
	buf = buf_new(2000);
#ifdef MSSH_RSA
	buf_readfile(buf, RSA_PRIV_FILENAME);
	buf_setpos(buf, 0);
	buf_get_priv_key(buf, hostkey);
#endif
#ifdef MSSH_DSS
	buf_setpos(buf, 0);
	buf_readfile(buf, DSS_PRIV_FILENAME);
	buf_setpos(buf, 0);
	buf_get_priv_key(buf, hostkey);
#endif
	buf_burn(buf);
	buf_free(buf);

	TRACE("leave loadhostkeys");
	return hostkey;
}
