#ifndef _OPTIONS_H_
#define _OPTIONS_H_

#include "includes.h"
#include "config.h"
#include "debug.h"

/******************************************************************
 * Define compile-time options below.
 ******************************************************************/

#define MSSH_PORT 2222 /* for testing */

/* Hostkey paths */
#define DSS_PRIV_FILENAME "mssh_dss_host_key"
#define RSA_PRIV_FILENAME "mssh_rsa_host_key"

/* Encryption - at least one required.
 * SSH2 RFC Draft requires 3DES and recommends blowfish, aes128 & twofish128 */
#define MSSH_AES128_CBC
#define MSSH_BLOWFISH_CBC
#define MSSH_TWOFISH128_CBC
#define MSSH_3DES_CBC

/* Integrity - at least one required.
 * SSH2 RFC Draft requires sha1-hmac, recommends md5-hmac */
#define MSSH_SHA1_HMAC
#define MSSH_MD5_HMAC

/* Hostkey/public key algorithms - at least one required, these are used
 * for hostkey as well as for verifying signatures with pubkey auth.
 * SSH2 RFC Draft requires dss, recommends rsa */
#define MSSH_RSA
#define MSSH_DSS

/* Define DSS_PROTOK to use Putty's method of generating the value k for dss,
 * rather than just from the random byte source.
 * Undefining this will save you ~4k in binary size with static uclibc, but
 * your DSS hostkey could be exposed if the random number source isn't good.
 * If in doubt, leave it defined */
#define DSS_PROTOK

/* Authentication type to use, at least one required.
   SSH2 RFC Draft requires pubkey auth, recommends password */
#define MSSH_PASSWORD_AUTH
#define MSSH_PUBKEY_AUTH

/* Random device to use - you must specify one only.
 * DEV_RANDOM is recommended on hosts with a good /dev/random
 * and /dev/urandom, otherwise use EGD and run EGD or PRNGD, specifying
 * the socket */
#define MSSH_DEV_RANDOM /* use /dev/random and /dev/urandom */

#undef MSSH_EGD /* use egd or prngd socket */
#define MSSH_EGD_SOCKET "./rng"

/* Specify the number of clients we will allow to be connected but
 * not yet authenticated. After this limit, connections are rejected */
#define MAX_UNAUTH_CLIENTS 30

/* The draft RFC recommends 20 tries, 5 seems more sensible */
#define MAX_AUTH_TRIES 5

/*******************************************************************
 * You shouldn't edit below here unless you know you need to.
 *******************************************************************/

#define MSSH_VERSION "0.1"
#define LOCAL_IDENT "SSH-2.0-mssh_" MSSH_VERSION

/* Timeouts in seconds */
#define SELECT_TIMEOUT 20
/* Spec recommends after one hour or 1 gigabyte of data */
#define KEX_REKEY_TIMEOUT 3600
#define KEX_REKEY_DATA (1<<30) /* 2^30 == 1GB, this value must be < INT_MAX */
/* Close connections to clients which haven't authorised after AUTH_TIMEOUT */
#define AUTH_TIMEOUT 600 /* 10 minutes is recommended */

#define MSSH_KEX_DH_GROUP1 0

#define MSSH_SIGNKEY_RSA 1
#define MSSH_SIGNKEY_DSS 0

/* Required for pubkey auth */
#ifdef MSSH_PUBKEY_AUTH
#define MSSH_SIGNKEY_VERIFY
#endif

#define MSSH_COMP_NONE 0
#define MSSH_COMP_ZLIB 1

/* SHA1 is 20 bytes == 160 bits */
#define SHA1_HASH_SIZE 20
/* SHA512 is 64 bytes == 512 bits */
#define SHA512_HASH_SIZE 64

/* 3DES requires a 24 byte key */
#define MAX_KEY_LEN 24
#define MAX_IV_LEN 20 /* must be same as max blocksize, 
						 and >= SHA1_HASH_SIZE */
#define MAX_MAC_KEY 20

#define MAX_NAME_LEN 64 /* maximum length of a protocol name, isn't
						   explicitly specified for all protocols (just
						   for algos) but seems valid */

#define MAX_PROPOSED_ALGO 20

/* size/count limits */
#define MAX_LISTEN_ADDR 10

#define MAX_PACKET_LEN 35000
#define MIN_PACKET_LEN 16
#define MAX_PAYLOAD_LEN 32768

#define MAX_TRANS_PAYLOAD_LEN 32768
#define MAX_TRANS_PACKET_LEN (MAX_TRANS_PAYLOAD_LEN+50)

#define DEV_RANDOM "/dev/urandom"
#define DEV_URANDOM "/dev/urandom"

#define _PATH_TTY "/dev/tty"

/* Due to the way libtomcrypt is compiled, we need to make sure these
 * things are explicitly enabled, since the makefile isn't exported.
 * These don't actually affect the compiled ciphers or the binary size */
#define CBC
#define DES
#define RIJNDAEL
#define BLOWFISH
#define TWOFISH
#define SHA1
#define MD5
#define SHA512
#define HMAC
#define MPI
#define SPRNG
#define BASE64

#endif /* _OPTIONS_H_ */
