#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>

#include "options.h"
#include "session.h"
#include "util.h"
#include "packet.h"
#include "algo.h"
#include "buffer.h"
#include "dss.h"
#include "ssh.h"
#include "random.h"
#include "kex.h"
#include "channel.h"


static void session_init(int sock, sign_key *hostkey, int childpipe);
static void session_identification();
static void checktimeouts();

struct sshsession ses;

void child_session(int sock, sign_key *hostkey, int childpipe) {

	fd_set readfd, writefd;
	struct timeval timeout;
	int val;
	
	initrandom();
	crypto_init();
	session_init(sock, hostkey, childpipe);

	/* exchange identification, version etc */
	session_identification();

	/* start off with key exchange */
	send_msg_kexinit();

	FD_ZERO(&readfd);
	FD_ZERO(&writefd);

	/* main loop, select()s for network and other connections */
	for(;;) {

		TRACE("top of select loop");
		timeout.tv_sec = SELECT_TIMEOUT; /* TODO */
		timeout.tv_usec = 0;
		FD_ZERO(&writefd);
		FD_ZERO(&readfd);
		assert(ses.payload == NULL);
		FD_SET(ses.sock, &readfd);
		if (!isempty(&ses.writequeue)) {
			FD_SET(ses.sock, &writefd);
		}

		/* set up for channels which require reading/writing */
		if (ses.dataallowed == 1) {
			setchannelfds(&readfd, &writefd);
		}
		val = select(ses.maxfd+1, &readfd, &writefd, NULL, &timeout);
#ifdef DEBUG_TRACE
		fprintf(stderr, "select val = %d\n", val);
#endif
		if (val < 0) {
			if (errno == EINTR) {
				continue;
			} else {
				mssh_error("error in select"); /* TODO - handle errors? */
			}
		}

		checktimeouts();
		
		if (val == 0) {
			/* timeout */
			TRACE("select timeout");
			continue;
		}


		if (FD_ISSET(ses.sock, &writefd) && !isempty(&ses.writequeue)) {
			write_packet();
		}

		if (FD_ISSET(ses.sock, &readfd)) {
			read_packet();
		}

		/* Process the decrypted packet. After this, the read buffer
		 * will be ready for a new packet */
		if (ses.payload != NULL) {
			process_packet();
		}

		/* process pipes etc for the channels */
		if (ses.dataallowed == 1) {
			channelio(&readfd, &writefd);
		}

	} /* for(;;) */
}

/* Check all timeouts which are required. Currently these are the time for
 * user authentication, and the automatic rekeying. */
void checktimeouts() {

	struct timeval tv;
	long secs;

	if (gettimeofday(&tv, 0) < 0) {
		mssh_error("error getting time");
	}

	secs = tv.tv_sec;
	
	if (!ses.authstate.authdone) {
		if (secs - ses.connecttime >= AUTH_TIMEOUT) {
			fprintf(stderr, "Timeout before userauth\n");
			exit(0);
		}
	}

	if (!ses.kexstate.sentkexinit
			&& (secs - ses.kexstate.lastkextime >= KEX_REKEY_TIMEOUT
			|| ses.kexstate.datarecv+ses.kexstate.datatrans >= KEX_REKEY_DATA)){
		TRACE("rekeying after timeout or max data reached");
		send_msg_kexinit();
	}
}


#ifdef DOCLEANUP
/* clean up session vars etc - only usually called at exit(), but useful
 * for checking memory left lying about */
void session_cleanup() {
	
	TRACE("enter session_cleanup");
	chancleanup();

	m_free(ses.session_id);
	sign_key_free(ses.hostkey);
	m_free(ses.remoteident);

	buf_free(ses.readbuf);
	buf_free(ses.decryptreadbuf);
	buf_free(ses.payload);
	buf_free(ses.writepayload);
	m_free(ses.keys);


	TRACE("leave session_cleanup");
}
#endif

/* called only at the start of a session, set up initial state */
static void session_init(int sock, sign_key *hostkey, int childpipe) {

	struct timeval tv;
	TRACE("enter session_init");

	ses.sock = sock;
	ses.maxfd = sock;

	ses.childpipe = childpipe;

	ses.hostkey = hostkey;

	if (gettimeofday(&tv, 0) < 0) {
		mssh_error("error getting time");
	}

	ses.connecttime = tv.tv_sec;
	
	kexinitialise(); /* initialise the kex state */
	authinitialise(); /* initialise auth state */
	chaninitialise(); /* initialise the channel state */

	ses.writepayload = buf_new(MAX_TRANS_PAYLOAD_LEN);
	ses.transseq = 0;

	ses.readbuf = NULL;
	ses.decryptreadbuf = NULL;
	ses.payload = NULL;
	ses.recvseq = 0;

	ses.expecting = SSH_MSG_KEXINIT;
	ses.dataallowed = 0; /* don't send data yet, we'll wait until after kex */

	/* set all the algos to none */
	ses.keys = (struct key_context*)m_malloc(sizeof(struct key_context));
	ses.newkeys = NULL;
	ses.keys->recv_algo_crypt = &mssh_nocipher;
	ses.keys->trans_algo_crypt = &mssh_nocipher;
	
	ses.keys->recv_algo_mac = &mssh_nohash;
	ses.keys->trans_algo_mac = &mssh_nohash;

	ses.keys->algo_kex = -1;
	ses.keys->algo_hostkey = -1;
	ses.keys->recv_algo_comp = MSSH_COMP_NONE;
	ses.keys->trans_algo_comp = MSSH_COMP_NONE;

#ifndef DISABLE_ZLIB
	ses.keys->recv_zstream = NULL;
	ses.keys->trans_zstream = NULL;
#endif

	/* key exchange buffers */
	ses.session_id = NULL;
	ses.kexhashbuf = NULL;
	ses.transkexinit = NULL;
	ses.dh_K = NULL;

	TRACE("leave session_init");
}

static void session_identification() {

	char linebuf[256];
	int len;
	int i;
	char done;

	/* write our version string, this blocks */
	writeln(ses.sock, LOCAL_IDENT "\r\n");

	/* now read the client version string, there are allowed to be other lines
	 * before the "SSH-*" line. We allow a max of 9 lines before it, just for
	 * sanity */
	done = 0;
	for (i = 0; i < 10; i++) {
		len = readln(ses.sock, linebuf, 256);
		if (len >= 4 && linebuf[0] == 'S'
					 && linebuf[1] == 'S'
					 && linebuf[2] == 'H'
					 && linebuf[3] == '-') {
			ses.remoteident = m_malloc(len+1);
			memcpy(ses.remoteident, linebuf, len);
			ses.remoteident[len] = '\0'; /* null terminated */
			done = 1;
			break;
		}
	}

	if (!done) {
		mssh_error("didn't get remote ident");
	}

	fprintf(stderr, "remoteident: %s\n", ses.remoteident);

}
