#include <stdlib.h>
#include <string.h>

#include "options.h"
#include "util.h"
#include "signkey.h"
#include "buffer.h"
#include "ssh.h"

/* malloc a new sign_key and set the dss and rsa keys to NULL */
sign_key * new_sign_key() {

	sign_key * ret;

	ret = (sign_key*)m_malloc(sizeof(sign_key));
#ifdef MSSH_DSS
	ret->dsskey = NULL;
#endif
#ifdef MSSH_RSA
	ret->rsakey = NULL;
#endif
	return ret;

}

void buf_get_pub_key(buffer *buf, sign_key *key) {

	unsigned char* ident;
	unsigned int len;

	ident = buf_getstring(buf, &len);

#ifdef MSSH_DSS
	if (memcmp(ident, SSH_SIGNKEY_DSS, len) == 0) {
		m_free(ident);
		buf_setpos(buf, buf->pos - len - 4);
		dss_key_free(key->dsskey);
		key->dsskey = (dss_key*)m_malloc(sizeof(dss_key));
		buf_get_dss_pub_key(buf, key->dsskey);
		return;
	}
#endif
#ifdef MSSH_RSA
	if (memcmp(ident, SSH_SIGNKEY_RSA, len) == 0) {
		m_free(ident);
		buf_setpos(buf, buf->pos - len - 4);
		rsa_key_free(key->rsakey);
		key->rsakey = (rsa_key*)m_malloc(sizeof(rsa_key));
		buf_get_rsa_pub_key(buf, key->rsakey);
		return;
	}
#endif

	m_free(ident);

	/* TODO this should be more forgiving */
	mssh_error("neither RSA nor DSS key was found");
	
}

void buf_get_priv_key(buffer *buf, sign_key *key) {

	unsigned char* ident;
	unsigned int len;

	ident = buf_getstring(buf, &len);

#ifdef MSSH_DSS
	if (memcmp(ident, SSH_SIGNKEY_DSS, len) == 0) {
		m_free(ident);
		buf_setpos(buf, buf->pos - len - 4);
		dss_key_free(key->dsskey);
		key->dsskey = (dss_key*)m_malloc(sizeof(dss_key));
		buf_get_dss_priv_key(buf, key->dsskey);
		return;
	}
#endif
#ifdef MSSH_RSA
	if (memcmp(ident, SSH_SIGNKEY_RSA, len) == 0) {
		m_free(ident);
		buf_setpos(buf, buf->pos - len - 4);
		rsa_key_free(key->rsakey);
		key->rsakey = (rsa_key*)m_malloc(sizeof(rsa_key));
		buf_get_rsa_priv_key(buf, key->rsakey);
		return;
	}
#endif

	m_free(ident);
	
	/* TODO this should be more forgiving */
	mssh_error("neither RSA nor DSS key was found");
	
}

/* signkeytype is either MSSH_SIGNKEY_DSS or MSSH_SIGNKEY_RSA */
void buf_put_pub_key(buffer* buf, sign_key *key, int signkeytype) {

	buffer *pubkeys;

	pubkeys = buf_new(1000);
	
#ifdef MSSH_DSS
	if (signkeytype == MSSH_SIGNKEY_DSS) {
		buf_put_dss_pub_key(pubkeys, key->dsskey);
	}
#endif
#ifdef MSSH_RSA
	if (signkeytype == MSSH_SIGNKEY_RSA) {
		buf_put_rsa_pub_key(pubkeys, key->rsakey);
	}
#endif
	if (pubkeys->len == 0) {
		mssh_error("bad key types in buf_put_pub_key");
	}

	buf_setpos(pubkeys, 0);
	buf_putstring(buf, buf_getptr(pubkeys, pubkeys->len),
			pubkeys->len);
	
	buf_free(pubkeys);
}

/* signkeytype is either MSSH_SIGNKEY_DSS or MSSH_SIGNKEY_RSA */
void buf_put_priv_key(buffer* buf, sign_key *key, int signkeytype) {

#ifdef MSSH_DSS
	if (signkeytype == MSSH_SIGNKEY_DSS) {
		buf_put_dss_priv_key(buf, key->dsskey);
	}
	return;
#endif
#ifdef MSSH_RSA
	if (signkeytype == MSSH_SIGNKEY_RSA) {
		buf_put_rsa_priv_key(buf, key->rsakey);
	}
	return;
#endif
	mssh_error("bad key types in put pub key");
}

void sign_key_free(sign_key *key) {

#ifdef MSSH_DSS
	dss_key_free(key->dsskey);
#endif
#ifdef MSSH_RSA
	rsa_key_free(key->rsakey);
#endif

	m_free(key);
}

void buf_put_sign(buffer* buf, sign_key *key, int signkeytype, 
		const unsigned char *data, unsigned int len) {

	buffer *sigblob;

	sigblob = buf_new(1000);

#ifdef MSSH_DSS
	if (signkeytype == MSSH_SIGNKEY_DSS) {
		buf_put_dss_sign(sigblob, key->dsskey, data, len);
	}
#endif
#ifdef MSSH_RSA
	if (signkeytype == MSSH_SIGNKEY_RSA) {
		buf_put_rsa_sign(sigblob, key->rsakey, data, len);
	}
#endif
	if (sigblob->len == 0) {
		mssh_error("non-matching signing type");
	}

	buf_setpos(sigblob, 0);
	buf_putstring(buf, buf_getptr(sigblob, sigblob->len),
			sigblob->len);
			
	buf_free(sigblob);

}

#ifdef MSSH_SIGNKEY_VERIFY
/* Return 1 if good signature, 0 otherwise. If 0 is returned, the position of
 * buf is undefined. If 1 is returned, buf will be positioned after the
 * signature blob */
int buf_verify(buffer * buf, sign_key *key, const unsigned char *data,
		unsigned int len) {
	
	unsigned int bloblen;
	unsigned char * ident = NULL;
	unsigned int identlen = 0;

	bloblen = buf_getint(buf);
	ident = buf_getstring(buf, &identlen);
#ifdef MSSH_DSS
	if (bloblen == DSS_SIGNATURE_SIZE &&
			memcmp(ident, SSH_SIGNKEY_DSS, identlen) == 0) {
		m_free(ident);
		return buf_dss_verify(buf, key->dsskey, data, len);
	}
#endif
#ifdef MSSH_RSA
	if (memcmp(ident, SSH_SIGNKEY_RSA, identlen) == 0) {
		m_free(ident);
		return buf_rsa_verify(buf, key->rsakey, data, len);
	}
#endif
	m_free(ident);
	mssh_error("non-matching signing type");
	return 0;
}
#endif /* MSSH_SIGNKEY_VERIFY */
